/*
 * List Implementation 
 *
 * $Id$
 *
 * Copyright (C) 2003-2005 Pengutronix, Hildesheim
 *
 * Written by Benedikt Spranger, Robert Schwebel
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include "ptxlist.h"

/*
 * 	Error strategy: in case of an error functions return NULL and set
 *	list_error_ptr to the current list entry. 
 */ 
ptxlist *list_error_ptr;


/*
 *	list_init - Constructor: create a new and empty list. 
 */
ptxlist *list_init (void)
{
	ptxlist *l;
	
	l = calloc(1, sizeof (ptxlist));
	if (!l){
		list_error_ptr = NULL;
		return NULL;
	}
	
	return l;
}

void goto_head (ptxlist **l)
{
	if (!l || !*l) return;
	while ((*l)->prev) *l = (*l)->prev;
}

void goto_tail (ptxlist **l)
{
	if (!l || !*l) return;
	while ((*l)->next) *l = (*l)->next;
}

void goto_next (ptxlist **l)
{
	if (!l || !*l) return;
	if ((*l)->next) *l = (*l)->next;
}

void goto_prev (ptxlist **l)
{
	if (!l || !*l) return;
	if ((*l)->prev) *l = (*l)->prev;
}

int is_last (ptxlist *l)
{
	if (!l) return -1;
	if (l->next) return 0;
	return 1;
}

int is_empty (ptxlist *l)
{
	if (!l) return -1; 
	if (!(l->data)) return 1;
	return 0;
}

int is_first (ptxlist *l)
{
	if (!l) return -1;
	if (l->prev) return 0;
	return 1;
}

/**
 *	list_add_after - add an element after the current element
 */
ptxlist *list_add_after (ptxlist *l, void *data, size_t size)
{
	ptxlist *tmp;
	char *d;

	if (!l) return NULL; 
	
	d = calloc(size, sizeof (char));
	if (!d) {
		list_error_ptr = l;
		return NULL;
	}

	/* empty lists have a ptxlist but no data */
	if (!is_empty(l)) {

		tmp = calloc(1, sizeof (ptxlist));
		if (!tmp) {
			free(d);
			list_error_ptr = l;
			return NULL;
		}

		tmp->next = l->next;
		tmp->prev = l;
		l->next = tmp;
	
	} else
		tmp = l; 
	
	memcpy(d, data, size);
	tmp->data = d;
	
	return tmp;
}

/**
 *	list_add_head - add an element to the beginning of the list
 */
ptxlist *list_add_head (ptxlist *l, void *data, size_t size)
{
	ptxlist *tmp;
	char *d;

	if (!l) return NULL;

	goto_head (&l);
	
	d = calloc(size, sizeof (char));
	if (!d) {
		list_error_ptr = l;
		return NULL;
	}

	if (!is_empty(l)) {	
		tmp = calloc(1, sizeof (ptxlist));
		if (!tmp) {
			free(d);
			list_error_ptr = l;
			return NULL;
		}

		tmp->next = l;
		l->next = tmp;
	
	} else
		tmp = l; 
	
	memcpy (d, data, size);
	tmp->data = d;	

	return tmp;
}


/**
 *	list_add_tail - add an element to the end of the list
 */
ptxlist *list_add_tail (ptxlist *l, void *data, size_t size)
{
	if (!l) return NULL;

	goto_tail(&l);
	
	return list_add_after(l, data, size); 
}


/*
 *	list_del - Destructor: free and destroy a list
 */
ptxlist *list_del (ptxlist *l)
{
	ptxlist *tmp;

	if (!l) return NULL;

	goto_head(&l);
	
	while (l) {
		tmp = l->next;
		free (l->data);
		free (l);
		l = tmp;
	}
	return NULL; 
}

/*
 *	list_del_element - remove one element from list
 */
ptxlist *list_del_element (ptxlist *l)
{
	ptxlist *tmp = NULL;

	if (!l) return NULL; 

	free(l->data);
	l->data=NULL; 

	if (l->next) {
		tmp = l->next;
		tmp->prev = l->prev;
		if (l->prev) l->prev->next=tmp;
		free(l);
		return tmp;
	}

	if (l->prev) {
		tmp = l->prev;
		tmp->next = NULL;
		free(l);
		return tmp;
	}

	return l; 
}

ptxlist *list_find (ptxlist *l, int (*fnct) (void *, void *), void *test)
{
	for (goto_head (&l); l;l = l->next)
	{
		if (fnct (l->data, test)) break;
	}
	return l;
}

void *list_get_element_data (ptxlist *l)
{
	if (!l) return NULL;
	return l->data;
}
